#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "general.h"

/*

midi2rtttl - Copyright 2002 Michael Kohn (mike@naken.cc)
This falls under the Kohnian license.  Please read
it at http://nakentone.naken.cc/

This program is NOT opensourced.  You may not use any part
of this program for your own software.

*/

int default_channel,default_track;
int header_flag;
int current_note;

struct _header_chunk
{
  int header_length;
  int format;
  int tracks;
  short int division;
};

struct _meta_events
{
  int sequence_number;
  int tempo;
  char smpte_offset[6];
  char time_signature[5];
  char key_signature[3];
  int quarter_note;
};

struct _rtttl_header
{
  int d;
  int o;
  int b;
};

int read_string(FILE *in, char *s, int n)
{
int t,ch;

  for(t=0; t<n; t++)
  {
    ch=getc(in);
    if (ch==EOF) return -1;

    s[t]=ch;
  }

  s[t]=0;
  return 0;
}

int print_string(FILE *in, int n)
{
int t,ch;

  for(t=0; t<n; t++)
  {
    ch=getc(in);
    if (ch==EOF) return -1;

    printf("%c",ch);
  }

  printf("\n");
  return 0;
}

int read_long(FILE *in)
{
int t;

  t=getc(in);
  t=(t<<8)+getc(in);
  t=(t<<8)+getc(in);
  t=(t<<8)+getc(in);

  return t;
}

int read_word(FILE *in)
{
int t;

  t=getc(in);
  t=(t<<8)+getc(in);

  return t;
}

int read_count(FILE *in, int n)
{
int t,s;

  t=0;

  for (s=0; s<n; s++)
  { t=(t<<8)+getc(in); }

  return t;
}

int read_var(FILE *in)
{
int t,ch;

  t=0;

  while(1)
  {
    ch=getc(in);

    if (ch==EOF) return -1;

    t=(t<<7)+(ch&127);
    if ((ch&128)==0) break;
  }

  return t;
}

int MThd_parse(FILE *in, struct _header_chunk *header_chunk, int n)
{
  header_chunk->header_length=n;
  header_chunk->format=read_word(in);
  header_chunk->tracks=read_word(in);
  header_chunk->division=read_word(in);

  return 0;
}

int parse_meta_event(FILE *in, struct _meta_events *meta_events, struct _header_chunk *header_chunk)
{
int ch,v_length,t;

  ch=getc(in);
  v_length=read_var(in);

  /* printf("(%d) ",ch); */

  if (ch==0x00)
  { 
    printf("Sequence Number: "); 
    t=read_count(in,v_length);
    meta_events->sequence_number=t;
    printf("%d\n",t);
    return 0;
  }
    else
  if (ch==0x01)
  { 
    printf("Text Event: "); 
    print_string(in, v_length);
    return 0;
  }
    else
  if (ch==0x02)
  { 
    printf("Copyright Notice: "); 
    print_string(in, v_length);
    return 0;
  }
    else
  if (ch==0x03)
  { 
    printf("Track Name: "); 
    print_string(in, v_length);
    return 0;
  }
    else
  if (ch==0x04)
  { 
    printf("Instrument Name: "); 
    print_string(in, v_length);
    return 0;
  }
    else
  if (ch==0x05)
  { 
    printf("Lyric Text: "); 
    print_string(in, v_length);
    return 0;
  }
    else
  if (ch==0x06)
  { 
    printf("Marker Text: "); 
    print_string(in, v_length);
    return 0;
  }
    else
  if (ch==0x07)
  { printf("Cue Point: "); }
    else
  if (ch==0x20)
  { printf("MIDI Channel Prefix: "); }
    else
  if (ch==0x2f)
  { 
#ifdef DEBUG
    printf("End Of Track\n"); 
#endif
    return -1;
  }
    else
  if (ch==0x51)
  { 
    printf("Tempo Setting: "); 
    t=read_count(in,v_length);
    meta_events->tempo=t;
    meta_events->quarter_note=meta_events->tempo/header_chunk->division;
    printf("%d\n",t);
    return 0;
  }
    else
  if (ch==0x54)
  { 
    printf("SMPTE Offset: ");
    read_string(in,meta_events->smpte_offset,5);
    for (t=0; t<5; t++) printf("%d ",meta_events->smpte_offset[t]);
    printf("\n");
    return 0;
  }
    else
  if (ch==0x58)
  {
    printf("Time Signature: ");
    read_string(in,meta_events->time_signature,4);
    for (t=0; t<4; t++) printf("%d ",meta_events->time_signature[t]);
    printf("\n");
    return 0;
  }
    else
  if (ch==0x59)
  {
    printf("Key Signature: ");
    read_string(in,meta_events->key_signature,2);
    for (t=0; t<2; t++) printf("%d ",meta_events->key_signature[t]);
    printf("\n");
    return 0;
  }
    else
  if (ch==0x7F)
  { printf("Sequencer Specific Event: "); }
    else
  { printf("\nUnknown Event: "); }

  for (t=0; t<v_length; t++)
  {
    ch=getc(in);
#ifdef DEBUG
    printf("%c ",ch);
#endif
  }

  printf("\n");

  return 0;
}

int parse_extras(FILE *in, int channel)
{
  if (channel==1)
  { getc(in); }
    else
  if (channel==2)
  { getc(in); getc(in); }
    else
  if (channel==3)
  { getc(in); }

  return 0;
}

int get_duration(int duration)
{
int t;
int duration_code[16]={ 0, 3125, 4687, 6250, 9375, 12500, 18750,
                        25000, 37500, 50000, 75000, 100000, 150000 };

  if (duration>150000) duration=150000;

  for (t=0; t<13; t++)
  {
    if (duration_code[t]==duration) return t-1;
    if (duration_code[t]<duration && duration_code[t+1]>duration)
    {
      if (duration-duration_code[t]<duration_code[t+1]-duration)
      { return t-1; }
        else
      { return t; }
    }
  }

  return t-1;
}
int add_note(FILE *out, int note, float duration, struct _rtttl_header *rtttl_header)
{
int o,d,t,f,k;

  /* d=(int)((duration/4)*100000); */
  d=(int)(duration*100000);
  d=get_duration(d);

  if (d==-1) return 0;

  if (header_flag==1)
  {
    header_flag=2;
  }
    else
  {
    fprintf(out,",");
  }

  k=d/2;
  t=1<<(5-k);
  if (!(d%2==0 && t==rtttl_header->d))
  { fprintf(out,"%d",t); }

#ifdef DEBUG
printf("> %d  %d < ",t,d%2);
#endif

  if (note==0)
  { fprintf(out,"p"); }
    else
  {
    t=note-57;
    if (t<0)
    {
      printf("Error: This song uses notes lower than RTTTL can do.\n");
      printf("Please use the -s option to raise the note to a higher scale./n");
      return -1;
    }

    f=t%12;
    o=t/12;

    if (f==0) fprintf(out,"a");
      else
    if (f==1) fprintf(out,"a#");
      else
    if (f==2) fprintf(out,"h");
      else
    if (f==3) fprintf(out,"c");
      else
    if (f==4) fprintf(out,"c#");
      else
    if (f==5) fprintf(out,"d");
      else
    if (f==6) fprintf(out,"d#");
      else
    if (f==7) fprintf(out,"e");
      else
    if (f==8) fprintf(out,"f");
      else
    if (f==9) fprintf(out,"f#");
      else
    if (f==10) fprintf(out,"g");
      else
    if (f==11) fprintf(out,"g#");

    o=o+5;
    if (o!=rtttl_header->o) fprintf(out,"%d",o);

    if (d%2==1) fprintf(out,".");
  }

  return 0;
}

int midi_command(FILE *in, FILE *out, int n, struct _rtttl_header *rtttl_header, struct _meta_events *meta_events, int v_time)
{
int command,channel;
int ch,note;
float duration;

  command=n&0xf0;
  channel=n&0x0f;

  if (command==0x80)
  {
    /* Sound Off */
    note=getc(in);
    ch=getc(in);

    if (v_time!=0)
    {
/*
      duration=((float)meta_events->tempo/(float)rtttl_header->b)/(float)v_time;
      duration=duration/((float)meta_events->quarter_note/100);
*/
      duration=(float)v_time/((float)meta_events->quarter_note*4);
#ifdef DEBUG
printf("[ %f %d ] ",duration,meta_events->quarter_note);
#endif
    }
      else
    { duration=0; }

    if (note!=current_note)
    {
      printf("Warning: Note to stop and previous note don't match.\n");
    }
      else
    {
      add_note(out, note, duration, rtttl_header);
    }
#ifdef DEBUG
printf("note: %d %f\n",note,duration);
#endif
  }
    else
  if (command==0x90)
  {
    /* Sound On */
    note=getc(in);
    ch=getc(in);

    if (v_time!=0)
    {
/*
      duration=((float)meta_events->tempo/(float)rtttl_header->b)/(float)v_time;
      duration=duration/((float)meta_events->quarter_note/100);
*/
      duration=(float)v_time/((float)meta_events->quarter_note*4);
#ifdef DEBUG
printf("[ %f %d ] ",duration,meta_events->quarter_note);
#endif
    }
      else
    { duration=0; }

    current_note=note;
    if (duration>=0.0001) add_note(out, 0, duration, rtttl_header);
#ifdef DEBUG
printf("note: %d %f\n",note,duration);
#endif
  }
    else
  if (command==0xa0)
  {
    /* Aftertouch */
    ch=getc(in);
    ch=getc(in);
  }
    else
  if (command==0xb0)
  {
    /* Controller */
    ch=getc(in);
    ch=getc(in);
  }
    else
  if (command==0xc0)
  {
    /* Program Change */
    ch=getc(in);
  }
    else
  if (command==0xd0)
  {
    /* Channel Pressure */
    ch=getc(in);
  }
    else
  if (command==0xe0)
  {
    /* Pitch Wheel */
    ch=getc(in);
    ch=getc(in);
  }
    else
  if (command==0xf0)
  {
    parse_extras(in,channel);
  }
    else
  {
    printf("Unknown MIDI code.  Please email Michael Kohn (mike@naken.cc)\n");
    printf("Code: ");
    print_hex(n);
    printf("\n");
  }

  return 0;
}

int MTrk_parse(FILE *in, FILE *out, struct _header_chunk *header_chunk, int n, struct _meta_events *meta_events)
{
int ch;
int v_time;
struct _rtttl_header rtttl_header;
#ifdef DEBUG
int x=0;
#endif

  rtttl_header.d=4;
  rtttl_header.o=5;
  rtttl_header.b=-1;

  printf("Length: %d\n",n);

  while (1)
  {
    v_time=read_var(in);

    ch=getc(in);

    if (ch==EOF)
    {
      printf("EOF found... oops.\n");
      return 0;
    }

    if (ch==0xff)
    {
      /* META Event */

#ifdef DEBUG
      printf("META Event\n");
      if (x!=0) { x=0; printf("\n"); }
#endif
      if (parse_meta_event(in,meta_events,header_chunk)==-1) break;
    }
      else
    if (ch==0xf0 || ch==0xf7)
    {
      /* System Exclusive Event */

#ifdef DEBUG
printf("System Exclusive\n");

      if (x!=0) { x=0; printf("\n"); }
#endif
      while(1)
      {
        ch=getc(in);
        printf("%c(%d)",ch,ch);
        if (ch==0xf7) break;
      }
    }
      else
    {
      /* MIDI DATA */

      if (header_flag==0)
      {
        rtttl_header.b=60000000/meta_events->tempo;
        fprintf(out,"d=%d,o=%d,b=%d:",rtttl_header.d,rtttl_header.o,rtttl_header.b);
        header_flag=1;
      }

#ifdef DEBUG
      printf("(%d) ",v_time);
      print_hex(ch);
      printf(" ");
#endif

      midi_command(in,out,ch,&rtttl_header,meta_events,v_time);

#ifdef DEBUG
      x++;
      if (x==20) { x=0; printf("\n"); }
#endif
    }
  }

#ifdef DEBUG
  printf("%d\n",meta_events->tempo);
#endif
  return 0;
}

int main (int argc, char *argv[])
{
FILE *in,*out;
char filename[1024];
char outname[1052];
char songname[32];
char header[5];
int n,t,track;
struct _header_chunk header_chunk;
struct _meta_events meta_events;

  default_channel=0;
  default_track=-1;
  songname[0]=0;
  outname[0]=0;
  header_flag=0;

  if (argc<2)
  {
    printf("\nmidi2rtttl "VERSION" - Convert MIDI 2 RTTTL.\n");
    printf(COPYRIGHT);
    printf("Usage: midi2rtttl [ options ] <input filename>\n");
    printf("     -c [ channel <default 0> ]\n");
    printf("     -t [ track <default all> ]\n");
    printf("     -o [ output filename ]\n");
    printf("     -n [ name of song ]\n");
    exit(0);
  }

  for (t=1; t<argc; t++)
  {
    if (strcmp(argv[t],"-c")==0)
    { 
      default_channel=atoi(argv[++t]);
    }
      else
    if (strcmp(argv[t],"-t")==0)
    { 
      default_track=atoi(argv[++t]);
    }
      else
    if (strcmp(argv[t],"-n")==0)
    { 
      strncpy(songname,argv[++t],10);
      songname[11]=0;
    }
      else
    if (strcmp(argv[t],"-o")==0)
    { 
      strncpy(outname,argv[++t],1023);
      songname[1023]=0;
    }
      else
    { strncpy(filename,argv[t],1023); }
  }

  if (outname[0]==0)
  {
    strcpy(outname,filename);
    t=strlen(outname);
    while(t>0 && outname[t]!='.') t--;
    if (t==0) t=strlen(outname);
    strcpy(&outname[t],".rtttl");
  }

  if (songname[0]==0)
  {
    t=strlen(outname);
    while(t>0 && outname[t]!='/' && outname[t]!='\\') t--;
    if (outname[t]=='/' || outname[t]=='\\') t++;
    for (n=0; n<10; n++)
    {
      if (outname[t]=='.' || outname[t]==0) break;
      songname[n]=outname[t++];
    }
    songname[n]=0; 
  }

  printf("Output File: %s\n",outname),
  printf("  Song Name: %s\n",songname),

  in=fopen(filename,"r");
  if (in==0)
  {
    printf("Could not open file for reading: %s\n",filename);
    exit(0);
  }

  out=fopen(outname,"w");
  if (out==0)
  {
    printf("Could not open file for writing: %s\n",outname);
    fclose(in);
    exit(0);
  }

  printf("\nmidi2rtttl "VERSION" (Beta) - Convert MIDI to RTTTL.\n");
  printf(COPYRIGHT);

  fprintf(out,"%s:",songname);

  track=0;

  meta_events.sequence_number=track;
  meta_events.tempo=500000;
  meta_events.time_signature[0]=4;
  meta_events.time_signature[1]=2;
  meta_events.time_signature[2]=0;
  meta_events.time_signature[3]=0;
  meta_events.key_signature[0]=0;
  meta_events.key_signature[1]=0;
  meta_events.quarter_note=5208;

  current_note=-1;

  while(1)
  {
    if (read_string(in,header,4)==-1)
    { break; }

    n=read_long(in);

    if (strcasecmp(header,"MThd")==0) 
    {
      MThd_parse(in,&header_chunk,n);
      printf("\n--------------------------------------------\n");
      printf("   Header length: %d\n",header_chunk.header_length);
      printf("          Format: %d\n",header_chunk.format);
      printf("Number of Tracks: %d\n",header_chunk.tracks);
      printf("        Division: %d\n",header_chunk.division);
      printf("--------------------------------------------\n");
      meta_events.quarter_note=meta_events.tempo/header_chunk.division;
    }
      else
    if (strcasecmp(header,"MTrk")==0) 
    {
      printf("\n--------------------------------------------\n");
      printf("Track: %d",track++);
      printf("\n--------------------------------------------\n");

      MTrk_parse(in, out, &header_chunk, n, &meta_events);
    }
  }

  fprintf(out,";");

  fclose(in);
  fclose(out);

  return (0);
}



