#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "general.h"

/*

rtttl2nokia - Copyright 2001 Michael Kohn (mike@naken.cc)
This falls under the Kohnian license.  Please read
it at http://nakentone.naken.cc/

This program is NOT opensourced.  You may not use any part
of this program for your own software.

*/

FILE *fp;
int pushback=-1;

int gettoken(char *token)
{
int tokentype;
int ptr,ch;

  tokentype=0;
  ptr=0;
  token[0]=0;

  while (1)
  {
    if (ptr>15) break;

    if (pushback!=-1)
    { ch=pushback; pushback=-1; }
      else
    { ch=getc(fp); }

    if ((ch==' ' || ch=='\t') || (ch=='\n' || ch=='\r')) continue;

    if (ch==EOF && ptr==0) return -1;
      else
    if (ch==EOF) { pushback=ch; break; }

    if ((ch>='A' && ch<='Z') || (ch>='a' && ch<='z'))
    {
      if (tokentype==1) { token[ptr++]=ch; }
        else
      if (ptr==0) { token[ptr++]=ch; tokentype=1; }
        else
      { pushback=ch; break; }
    }
      else
    if (ch>='0' && ch<='9')
    {
      if (tokentype==2) { token[ptr++]=ch; }
        else
      if (ptr==0) { token[ptr++]=ch; tokentype=2; }
        else
      { pushback=ch; break; }
    }
      else
    {
      if (ptr==0)
      {
        tokentype=3;
        token[ptr++]=ch;
        break;
      }
        else
      {
        pushback=ch;
        break;
      }
    }

  }

  token[ptr]=0;
  return tokentype;
}

int main (int argc, char *argv[])
{
char filename[1024];
char songname[16];
char token[16];
int t,r,l,ch;
int d,o,b,scale,count,a0,b0;
int length,tone,octave,modifier;
int higher,looping,style;
int warning,full_sms_size,a440;
int headers,volume;

  d=4;
  o=6;
  b=63;

  count=0;
  stackptr=0;
  bitptr=7;
  scale=-1;
  higher=0;
  looping=0;
  style=0;
  warning=0;
  full_sms_size=SMS_SIZE;
  a440=5;
  headers=1;
  volume=VOLUME;

  if (argc<2)
  {
    printf("\nrtttl2nokia "VERSION" - Convert RTTTL files to Nokia.\n");
    printf(COPYRIGHT);
    printf("Usage: rtttl2nokia [options ] <input filename>\n");
    printf("       -o [ Make song 1 octave higher ]\n");
    printf("       -x [ remove all headers ]\n");
    printf("       -u [ only user-data-header (8bit) ]\n");
    printf("       -k [ include user-data-header and //SCKL ]\n");
    printf("       -r [ Set A440 as octave 4 (rtttl spec is octave 5) ]\n");
    printf("       -s <sms size>  [ Maximum size of SMS message (default %d) ]\n",SMS_SIZE);
    printf("       -v <volume>    [ (1 to 15) (default %d) ]\n\n",VOLUME);
    exit(0);
  }

  for (t=1; t<argc; t++)
  {
    if (strcmp(argv[t],"-o")==0)
    { higher=1; }
      else
    if (strcmp(argv[t],"-r")==0)
    { a440=4; }
      else
    if (strcmp(argv[t],"-x")==0)
    { headers=0; }
      else
    if (strcmp(argv[t],"-k")==0)
    { headers=2; }
      else
    if (strcmp(argv[t],"-u")==0)
    { headers=3; }
      else
    if (strcmp(argv[t],"-v")==0)
    { volume=atoi(argv[++t]); }
      else
    if (strcmp(argv[t],"-s")==0)
    { full_sms_size=atoi(argv[++t]); }
      else
    { strncpy(filename,argv[t],1023); }
  }

  fp=fopen(filename,"r");
  if (fp==0)
  {
    printf("Could not open file for reading: %s\n",filename);
    exit(0);
  }

  push(2,8);   /* ring tone commands */
  push(74,8);
  push(29,7);

  l=0;
  while(1)
  {
    ch=getc(fp);
    if (ch==':') break;
    if (ch==EOF)
    {
      printf("Error: Illegal RTTTL format file\n");
      exit(0);
    }

    if (l<15) songname[l++]=ch;
  }

  songname[l]=0;
/*
  if (gettoken(songname)==-1)
  {
    printf("Error: Illegal RTTTL format file\n");
    exit(0);
  }

  l=strlen(songname);
*/


  push(1,3);   /* song type */
  push(l,4);   /* length of song name */

  printf("\nrtttl2nokia "VERSION" - Convert RTTTL files to Nokia.\n");
  printf(COPYRIGHT);

  printf("\n       Song Name: %s\n",songname);

  for (t=0; t<l; t++)
  { push(songname[t],8); }

  push(1,8);   /* 1 song pattern */
  push(0,3);   /* pattern header id */
  push(0,2);   /* A-part */
#ifdef DEBUG
  printf("stackptr=%d   bitptr=%d\n",stackptr,bitptr);
#endif

  /* gettoken(token); */

  r=0;
  while(1)
  {
    if (gettoken(token)==-1)
    {
      printf("Error: Illegal RTTTL format file\n");
      exit(0);
    }

    if (r==0)
    {
      if (strcmp(token,",")==0) continue;
        else
      if (strcmp(token,":")==0) break;
        else
      if (strcasecmp(token,"d")==0) r=1;
        else
      if (strcasecmp(token,"o")==0) r=2;
        else
      if (strcasecmp(token,"b")==0) r=3;
        else
      if (strcasecmp(token,"s")==0) r=4;
        else
      if (strcasecmp(token,"l")==0) r=5;
    }
      else
    if (strcmp(token,"=")==0)
    {
      gettoken(token);

      if (r==1) d=atoi(token);
        else
      if (r==2) o=atoi(token);
        else
      if (r==3) b=atoi(token);
        else
      if (r==4)
      {
        if (strcasecmp(token,"n")==0)
        { style=0; }
          else
        if (strcasecmp(token,"c")==0)
        { style=1; }
          else
        if (strcasecmp(token,"s")==0)
        { style=2; }
      }
        else
      if (r==5) looping=atoi(token);
      r=0;
    }
      else
    {
      printf("Error: Illegal RTTTL format file\n");
      exit(0);
    }
  }

  if (looping==15) looping=31;

  printf("Default Duration: %d\n",d);
  printf("   Default Scale: %d\n",o);
  printf("Beats Per Minute: %d\n",b);
  printf("           Style: ");
  if (style==0) printf("Natural\n");
    else
  if (style==1) printf("Continuous\n");
    else
  if (style==2) printf("Staccato\n");
  printf("            Loop: %d\n",looping);
  printf("        SMS Size: %d\n",full_sms_size);

#ifdef DEBUG
  printf("stackptr=%d   bitptr=%d\n",stackptr,bitptr);
#endif

  push(looping,4); /* looping  */

  a0=stackptr;
  b0=bitptr;
  push(0,8);

  t=get_tempo(b);
  push(4,3);       /* set tempo */
  push(t,5);
  count++;

  push(3,3);       /* set style */
  push(style,2);
  count++;

  push(5,3);       /* set volume */
  push(volume,4);
  count++;

  while(1)
  {
    t=gettoken(token);
    if (t==-1) 
    { 
#ifdef DEBUG
printf("EOF\n");
#endif
      break; 
    } 

#ifdef DEBUG
    printf("%s\n",token);
#endif
    if (strcmp(token,";")==0) break; 

    length=d;

    if (t==2) 
    {
      length=atoi(token);
      t=gettoken(token);
    }

#ifdef DEBUG
    printf("Length: %d\n",length);
#endif

    if (length==1) length=0;
      else
    if (length==2) length=1;
      else
    if (length==4) length=2;
      else
    if (length==8) length=3;
      else
    if (length==16) length=4;
      else
    if (length==32) length=5;

    token[0]=tolower(token[0]);

    if (token[0]=='p') tone=0;
      else
    if (token[0]=='a') tone=10;
      else
    if (token[0]=='h' || token[0]=='b') tone=12;
      else
    if (token[0]=='c') tone=1;
      else
    if (token[0]=='d') tone=3;
      else
    if (token[0]=='e') tone=5;
      else
    if (token[0]=='f') tone=6;
      else
    if (token[0]=='g') tone=8;
      else
    if (token[0]=='o') tone=1;
      else
    if (token[0]=='s') tone=2;

#ifdef DEBUG
  printf("tone: %d\n",tone);
#endif

    t=gettoken(token);

    if (token[0]=='=')
    {
#ifdef DEBUG
printf("equals\n");
#endif
      t=gettoken(token);
#ifdef DEBUG
printf("token: %s\n",token);
#endif

      if (tone==12)
      {
        b=atoi(token);
        t=get_tempo(b);
        push(4,3);       /* set tempo */
        push(t,5);
        count++;
      }
        else
      if (tone==1)
      {
        o=atoi(token);
      }
        else
      if (tone==2)
      {
        if (strcasecmp(token,"n")==0)
        { style=0; }
          else
        if (strcasecmp(token,"c")==0)
        { style=1; }
          else
        if (strcasecmp(token,"s")==0)
        { style=2; }

        push(3,3);       /* set style */
        push(style,2);
        count++;
      }

      t=gettoken(token);
      if (strcmp(token,",")!=0)
      {
        printf("Error: expecting , and found %s\n",token);
      } 

      continue;
    }

    modifier=0;
    octave=-1;

    while(1)
    {
      if (token[0]==',' || token[0]==';' || t==-1) break;

      if (token[0]=='#')
      { tone=tone+1; }
        else
      if (t==2)
      { octave=atoi(token); }
        else
      if (strcmp(token,".")==0)
      { modifier++; }

      t=gettoken(token);
    }

    if (tone!=0)
    {
      if (octave==-1) octave=o;

      if (scale!=octave)
      {
        if (octave==a440) warning++;
        scale=octave;
        push(2,3);
        push(octave-a440+higher,2);
        count++;
      }
    }

    push(1,3);
    push(tone,4);
    push(length,3);
    push(modifier,2);
    count++;

/*
    if (scale!=octave)
    {
      scale=octave;
      push(2,3);
      push(octave-5,2);
      count++;
    }
*/

    if (t==-1)
    { break; }
  }

  if (bitptr!=7) stackptr++;
  push(0,8);

#ifdef DEBUG
  printf("commands: %d    a0=%d   b0=%d\n",count,a0,b0);
#endif

  push_addr(count,8,a0,b0); 

  print_codes(full_sms_size,"1581",headers);

  if (warning!=0 && higher==0)
  {
    printf("Warning: Song uses lower octave.  Please read docs on the -o option.\n");
  }

  printf("\n");

  return (0);
}



